﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using UnityEngine.Rendering;
using static AQUAS_Parameters;

/*
	DLUtils plugin can:
    Edit builtin assets or CUA: 
        - change scale(X,Y,Z)
        - change game object layer (for all renderers together or per each one)
        - override anchor (because of reflection/light probe)
        - switch for "cast shadows" and for "receive shadows"
    Edit material (for each material in asset) :
        - turn off GameObject
        - turn off mesh renderers (per each one)
        - change main color
        - change main alpha (opacity)
        - change emission color (if shader supports)
        - change stencil reference (if shader supports)
        - change render queue (only for all materials at once)
        - set render mode (if shader supports)
    Edit light (first found):
        - set culling mask (list of layers the lihgt interract width)

    For portal system, if scene has CUA with special pack of shaders named "cubeshaders",
    on start, plugin will try replace a shader for each material on stencil able ones, from pack.

	Made by DemonLord
	License: Creative Commons with Attribution (CC BY)
	Based on UnityAssetVamifier code from NoStage3
*/

namespace DLPlugin {
    public class DLUtils : MVRScript {
        private readonly string VERSION_NUMBER = "1.1";

        protected UIDynamicPopup uipop32;
        protected UIDynamicSlider setAlpha;
        protected UIDynamicToggle uiRendEnab;
        protected UIDynamicToggle uiGOEnab;

        protected List<UIDynamicColorPicker> UIColorList = new List<UIDynamicColorPicker>();

        protected JSONStorableStringChooser jpop;
        protected JSONStorableStringChooser jpop32;
        protected JSONStorableStringChooser jpopStencilRef;
        protected JSONStorableStringChooser jpopStencilRef2;
        protected JSONStorableStringChooser jpopSelectTransform;
        protected JSONStorableStringChooser jpopRenderMode;
        protected JSONStorableStringChooser jpopCubeShaders;
        protected JSONStorableString infoBox;
        protected JSONStorableFloat jsScaleX;
        protected JSONStorableFloat jsScaleY;
        protected JSONStorableFloat jsScaleZ;
        protected JSONStorableFloat l1CullingMask;
        protected JSONStorableFloat jsRenderQueue;
        protected JSONStorableBool jCastShadows;
        protected JSONStorableBool jReceiveShadows;
        protected JSONStorableBool jShadowsZWrite;
        protected JSONStorableBool jStopEditMaterial;

        protected List<JSONStorableStringChooser> jpop32List = new List<JSONStorableStringChooser>();
        protected List<JSONStorableColor> ColorsList = new List<JSONStorableColor>();
        protected List<JSONStorableFloat> AlphasList = new List<JSONStorableFloat>();
        protected List<JSONStorableBool> RendEnabList = new List<JSONStorableBool>();
        protected List<JSONStorableBool> layersList = new List<JSONStorableBool>();
        protected List<JSONStorableBool> GOEnabList = new List<JSONStorableBool>();

        private List<string> layers32 = new List<string>();
        private List<string> renderersNames = new List<string>();
        private List<MatPackage> origMaterials = new List<MatPackage>();
        private List<MatPackage> newMaterials = new List<MatPackage>();
        private List<Renderer> allRenderers = new List<Renderer>();
        private List<Renderer> allRenderersWas = new List<Renderer>();
        private CustomUnityAssetLoader cUAL;
        private CustomUnityAssetLoader cUAL2;
        private List<Renderer> allRenderers2 = new List<Renderer>();

        private JSONNode node;
        private JSONNode myJsnode;
        private Light l1;
        private int l1MaskWas;
        private bool assetChanged;
        private Atom shadersAtom;
        private bool haveStencilMask;
        private bool useAllRenderers;
        private readonly List<string> blendMode = new List<string> { "Opaque", "Cutout", "Fade", "Transparent" };
        private Transform rescaleObject;
        private Vector3 scale;
        private Vector3 scaleWas;

        public struct MatPackage {
            public int rendererIndex;
            public int materialIndex;
            public Material material;
            public bool isConverted;
            public string colorName;

            public MatPackage(int rIndex, int mIndex, Material mat, bool converted) {
                this.rendererIndex = rIndex;
                this.materialIndex = mIndex;
                this.material = mat;
                this.isConverted = converted;
                this.colorName = "R" + rIndex + "M" + mIndex;
            }
        }

        public override void Init() {
            try
            {
                renderersNames.Add("ALL");
                for (int i = 0; i < 32; i++)
                {
                    layers32.Add(i.ToString());
                }

                myJsnode = new JSONNode();
                if (!assetChanged && (SuperController.singleton.loadJson?["atoms"] ?? null) != null)
                {
                    foreach (JSONNode a in SuperController.singleton.loadJson["atoms"].AsArray)
                    {
                        if (a["id"].Value == containingAtom.name)
                        {
                            foreach (JSONNode n in a["storables"].AsArray)
                            {
                                if (n["id"].Value == this.name)
                                {
                                    //SuperController.LogMessage("Init myJsnode: " + n.Count);
                                    myJsnode = n;
                                    //GetLoadNamesM();
                                    break;
                                }
                            }
                        }
                    }
                }

                for (int i = 0; i < containingAtom.childAtomContainer.childCount; i++)
                {
                    if (containingAtom.childAtomContainer.GetChild(i).name == "rescaleObject")
                    {
                        rescaleObject = containingAtom.childAtomContainer.GetChild(i);
                        scale = new Vector3(rescaleObject.transform.localScale.x, rescaleObject.transform.localScale.y, rescaleObject.transform.localScale.z);
                        scaleWas = scale;/*
                        SuperController.LogMessage("rescaleObject.childCount " + rescaleObject.childCount);
                        if (rescaleObject.childCount > 1)
                        {
                            int c = rescaleObject.childCount;
                            for (int u = 0; u < c; u++)
                            {
                                SuperController.LogMessage("name " + rescaleObject.GetChild(0).name);
                                rescaleObject.GetChild(u).position = rescaleObject.transform.position;
                            }
                        }*/
                        break;
                    }
                }
                if (SuperController.singleton != null)
                {
                    SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename1;
                }
                l1 = containingAtom.childAtomContainer.GetComponentInChildren<Light>(true);

                BuildUIControls();

                AddVersionNumberToUI();
                /*
                foreach (Atom a in SuperController.singleton.GetAtoms())
                {
                    //if (a.name.StartsWith("1"))ShowName(a.transform, a.transform.name);
                    //SuperController.LogMessage("find " + a.transform.name);
                    *//*for (int i = 0; i < a.childAtomContainer.childCount; i++)
                    {
                        ShowName(a.childAtomContainer.GetChild(i), "");
                    }*//*
                    //SuperController.LogMessage("find "+ a.name);
                    if (a.name.ToLower() == "portalmain")
                    {
                        portalAtom = a;
                        SuperController.LogMessage("find portalAtom");

                    }
                }*/
                /*
                AssetLoader.AssetBundleFromFileRequest req = new AssetLoader.AssetBundleFromFileRequest()
                {
                    path = GetPackagePath(this) + "Custom/Assets/DemonLord/custom_subsurface_glossnmtessmappedfixedcomputebuff.assetbundle",
                    callback = LoadAssetBundle,
                };
                AssetLoader.QueueLoadAssetBundleFromFile(req);*/

                foreach (Atom a in SuperController.singleton.GetAtoms())
                {
                    //SuperController.LogMessage(a.name);
                    if (a.name.ToLower() == "cubeshaders")
                    {
                        shadersAtom = a;
                        //SuperController.LogMessage("found shadersAtom");
                        cUAL2 = shadersAtom.GetComponentInChildren<CustomUnityAssetLoader>();
                        cUAL2.RegisterAssetLoadedCallback(AssetLoaded2);
                        cUAL2.RegisterAssetClearedCallback(AssetCleared2);
                        if (cUAL2.isAssetLoaded)
                        {
                            AssetLoaded2();
                        }
                        break;
                    }
                }
                if (shadersAtom == null)
                {
                    cUAL = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
                    // get the loader component
                    SetCUACallback(cUAL);
                }
                /*
                if (containingAtom.type == "Person")
                {
                    foreach (var item in containingAtom.GetComponentsInChildren<Renderer>())
                    {
                        SuperController.LogMessage("AssetLoaded2 Renderer " + item.name + " \tMat " + item.material.name + " \tShader " + item.material.shader.name);
                    }
                    skin = containingAtom.GetComponentInChildren<DAZCharacterSelector>().selectedCharacter.skin;
                    foreach (var item in skin.GPUmaterials)
                    {
                        SuperController.LogMessage("AssetLoaded2 Mat " + item.name + " \tShader " + item.shader.name);
                    }
                    foreach (var item in containingAtom.GetComponentInChildren<DAZCharacterSelector>().hairItems)
                    {
                        if (item.active) SuperController.LogMessage("AssetLoaded2 Hair " + item.name + item.GetComponentInChildren<Renderer>().material.shader.name);
                    }
                    foreach (var item in containingAtom.GetComponentInChildren<DAZCharacterSelector>().clothingItems)
                    {
                        if (item.active)
                        {
                            SuperController.LogMessage("AssetLoaded2 Clothing " + item.name);
                            foreach (var item2 in item.GetComponentsInChildren<DAZMesh>())
                            {
                                SuperController.LogMessage("    AssetLoaded2 DAZMesh " + item2.name);
                                foreach (var item3 in item2.materials)
                                {
                                    SuperController.LogMessage("        AssetLoaded2 shader " + item3.shasder.name);
                                }
                            }
                        }
                    }
                    //SuperController.LogMessage("skinMethod " + skin.skinMethod.ToString());
                    //skin.GPUmaterials[0].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
                    //JSONStorableFloat jsfA = new JSONStorableFloat("ALPHA", 1, SetALPHA, 0, 1, false);
                    //CreateSlider(jsfA);
                    //skin.BroadcastMessage("OnApplicationFocus", true);
                }*/


                // since there is no "asset bundle loaded" callback yet, we have to wait
                //StartCoroutine(WaitForAssetBundle());
                if (cUAL == null)
                    containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
                //SuperController.LogMessage("1 allRenderers: " + allRenderers.Count);
                //SuperController.LogMessage("1 allRenderers2: " + allRenderers2.Count);
                if (cUAL == null && allRenderers.Count> 0 && shadersAtom == null)
                    ConvertMaterial();

            }
            catch (Exception e) {
                SuperController.LogError("DLUtils: " + e);
                //SuperController.LogError("DLUtils: " + this.transform.parent.parent.parent.parent.parent.name + " " + e);
            }
        }
        /*
        public static string GetPackagePath(MVRScript self)
        {
            char[] PathSeparatorChars = new char[] { '/', '\\' };
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            filename = filename.Substring(0, filename.LastIndexOfAny(PathSeparatorChars));
            int idx = filename.IndexOf(":/");
            if (idx >= 0)
                return filename.Substring(0, idx + 2);
            else
                return string.Empty;
        }
       
        private void LoadAssetBundle(MeshVR.AssetLoader.AssetBundleFromFileRequest request)
        {
            if (request.assetBundle == null)
            {
                SuperController.LogMessage("LoadAssetBundle null");
                return;
            }
          
            string path = "Assets/Custom/Shaders/";
            myShader1 = request.assetBundle.LoadAsset<Shader>(path + "Custom_Subsurface_GlossNMTessMappedFixedComputeBuff.shader");
            SuperController.LogMessage("LoadAssetBundle "+ (myShader1==null));
        }*/
        protected void SetCUACallback(CustomUnityAssetLoader cUAL)
        {
            if (cUAL != null)
            {
                cUAL.RegisterAssetClearedCallback(AssetCleared);
                cUAL.RegisterAssetLoadedCallback(AssetLoaded);
                if (cUAL.isAssetLoaded)
                {
                    AssetLoaded();
                }
            }
        }

        protected void BuildUIControls() {
            jpop = new JSONStorableStringChooser("Objects", renderersNames, "ALL", "Objects", Setpop);
            CreateScrollablePopup(jpop);

            jpopRenderMode = new JSONStorableStringChooser("RenderMode", blendMode, "None", "Render Mode", SetRenderMode);
            RegisterStringChooser(jpopRenderMode);

            jpop32 = new JSONStorableStringChooser("ALL Layer", new List<string>(layers32), "0", "ALL Layer", Setpop32);
            RegisterStringChooser(jpop32);
            uipop32 = CreateScrollablePopup(jpop32);

            jpopStencilRef = new JSONStorableStringChooser("StencilRef", new List<string>(layers32), "0", "StencilRef", SetpopStencilRef);
            RegisterStringChooser(jpopStencilRef);
            jpopStencilRef2 = new JSONStorableStringChooser("StencilRef2", new List<string>(layers32), "0", "StencilRef2", SetpopStencilRef);
            RegisterStringChooser(jpopStencilRef2);

            jCastShadows = new JSONStorableBool("CastShadows", true, SetCastShadows);
            RegisterBool(jCastShadows);
            jReceiveShadows = new JSONStorableBool("ReceiveShadows", true, SetReceiveShadows);
            RegisterBool(jReceiveShadows);

            jStopEditMaterial = new JSONStorableBool("StopEditMaterial", false, SetStopEditMaterial );
            RegisterBool(jStopEditMaterial);


            List<string> s = new List<string>();
            foreach (Atom a in SuperController.singleton.GetAtoms())
            {
                s.Add(a.name);
            }
            jpopSelectTransform = new JSONStorableStringChooser("Anchor Override", s, containingAtom.name, "Anchor Override", SetAnchor);
            RegisterStringChooser(jpopSelectTransform);


            if (l1 != null) {
                for (int i = 0; i < 32; i++)
                {
                    layersList.Add(new JSONStorableBool("Layer " + i.ToString(), false));
                    RegisterBool(layersList[i]);
                    layersList[i].isStorable = layersList[i].isRestorable = false;
                }
                l1MaskWas = l1.cullingMask;
                renderersNames.Add("LIGHT");
                l1CullingMask = new JSONStorableFloat("l1CullingMask", (float)l1MaskWas, -2147483648f, 2147483647f, true);
                RegisterFloat(l1CullingMask);
                if (GetStoredNode("l1CullingMask", out node))
                    l1CullingMask.val = node.AsFloat;
                //SuperController.LogMessage("light2: " + l1.cullingMask + " l1CullingMask: " + l1CullingMask.val);
                char[] stringArray = Convert.ToString((int)Math.Round(l1CullingMask.val), 2).ToCharArray();
                Array.Reverse(stringArray);
                for (int i = 0; i < 32; i++) {
                    if (stringArray.Length <= i)
                        layersList[i].val = false;
                    else if (stringArray[i] == '0')
                        layersList[i].val = false;
                    else
                        layersList[i].val = true;
                    layersList[i].setJSONCallbackFunction += SetLayerN;
                }
                l1.cullingMask = (Int32)l1CullingMask.val;
            }

            // scaleX 
            jsScaleX = new JSONStorableFloat("scaleX", rescaleObject.transform.localScale.x, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleX);
            jsScaleY = new JSONStorableFloat("scaleY", rescaleObject.transform.localScale.y, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleY);
            jsScaleZ = new JSONStorableFloat("scaleZ", rescaleObject.transform.localScale.z, SetScaleXYZ, 0f, 10f, false);
            RegisterFloat(jsScaleZ);

            if (!useAllRenderers) Setpop(jpop.val);

            infoBox = new JSONStorableString("info", "\ninfo");
            UIDynamicTextField infoTextField = CreateTextField(infoBox, true);
            infoTextField.height = 520;
        }


        protected void SetScaleXYZ(JSONStorableFloat jf)
        {
            if (jf.name == "scaleX") scale.x = jf.val;
            if (jf.name == "scaleY") scale.y = jf.val;
            if (jf.name == "scaleZ") scale.z = jf.val;
            rescaleObject.transform.localScale = scale;
            //SuperController.LogMessage("scale "+x1+" "+y1+" "+z1);
        }
        protected void LayerSet(Transform t, int ix = 0)
        {
            t.gameObject.layer = ix;
            for (int i = 0; i < t.childCount; i++)
            {
                LayerSet(t.GetChild(i), ix);
            }
        }/*
        protected void ShowName(Transform t, string tabs)
        {
            tabs += "  ";
            SuperController.LogMessage(tabs + "name : " + t.name + " pos: " + t.transform.position);
            //SuperController.LogMessage("Plane " + t.transform.position);
            for (int i = 0; i < t.childCount; i++)
            {
                ShowName(t.GetChild(i), tabs);
            }
        }*/
        protected void AddVersionNumberToUI() {
            JSONStorableString versionString = new JSONStorableString("pluginVersion", "\nDLUtils " + VERSION_NUMBER + "\nby DemonLord (CC BY)");
            UIDynamicTextField textFieldVersion = CreateTextField(versionString, true);
            textFieldVersion.height = 120;
        }


        protected void AssetCleared()
        {
            assetChanged = true;
            foreach (JSONStorableColor colorN in ColorsList) DeregisterColor(colorN);
            foreach (JSONStorableStringChooser chooser in jpop32List) DeregisterStringChooser(chooser);

            allRenderers.Clear();
            origMaterials.Clear();
            newMaterials.Clear();
            renderersNames.Clear();
            renderersNames.Add("ALL");
            jpop32List.Clear();
            ColorsList.Clear();
            //SuperController.LogMessage("Clear "+ allRenderers.Count);
        }

        protected void AssetCleared2()
        {
            RestoreOriginalMaterials();
            AssetCleared();
            allRenderers2.Clear();
            AssetLoaded();
        }

        protected void AssetLoaded()
        {
            containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
            ConvertMaterial();
        }
        protected void AssetLoaded2()
        {
            shadersAtom.GetComponentsInChildren<Renderer>(false, allRenderers2);
            cUAL = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
            // get the loader component

            if (cUAL == null)
                containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
            if (cUAL == null && allRenderers.Count > 0) {
                ConvertMaterial();
                return;
            }
            SetCUACallback(cUAL);
        }

        protected void SetFirstVal() {
            if (containingAtom.childAtomContainer.childCount > 0) {
                jpop32.val = containingAtom.childAtomContainer.GetChild(0).gameObject.layer.ToString();
            }
            //SuperController.LogMessage("jpop32 val: "+ jpop32.val);
        }
        protected void ConvertMaterial() {
            useAllRenderers = true;
            jpop32List.Clear();
            //SuperController.LogMessage("allRenderers Count: "+ allRenderers.Count);
            for (int i = 0; i < allRenderersWas.Count; i++)
            {
                allRenderers.Remove(allRenderersWas[i]);
            }
            allRenderersWas.Clear();
            //SuperController.LogMessage("allRenderers Count2: " + allRenderers.Count);
            if (allRenderers.Count == 0) {
                SuperController.LogMessage("DLUtils: Make sure you have an .assetbundle loaded and a prop selected from the Asset dropdown. Then reload this plugin.");
                return;
            }

            if (GetStoredNode("StopEditMaterial", out node))
                jStopEditMaterial.val = node.AsBool;
            if (jStopEditMaterial.val) allRenderers.Clear();

            for (int i = 0; i < allRenderers.Count; i++) {
                allRenderersWas.Add(allRenderers[i]);//bug load previous CUA renderers, this will remember them to delete on next CUA
                jpop32List.Add(new JSONStorableStringChooser("Render " + i + " Layer", new List<string>(layers32), "0", "Layer", Setpop32N));
                GOEnabList.Add(new JSONStorableBool("GameObject " + i, allRenderers[i].gameObject.activeSelf, SetGOEnable));
                RendEnabList.Add(new JSONStorableBool("Render " + i, allRenderers[i].enabled, SetRenderEnable));
                for (int j = 0; j < allRenderers[i].materials.Length; j++)
                {
                    renderersNames.Add("Render " + i + " " + "Mat " + j);
                    Material mat = Material.Instantiate(allRenderers[i].materials[j]);
                    //SuperController.LogMessage("shader.name " + allRenderers[i].materials[j].shader.name);

                    if (allRenderers[i].materials[j].shader.name == "Custom/StencilMask" || allRenderers[i].materials[j].shader.name == "MAG/Blackhole")
                    {
                        haveStencilMask = true;
                    }
                    else if (!mat.shader.name.Contains("AQUAS"))//dont change water material
                    {
                        if (allRenderers2.Count > 0)
                        {
                            for (int r = 0; r < allRenderers2.Count; r++)
                            {
                                Shader s = allRenderers2[r].material.shader;
                                if (s.name.Contains(mat.shader.name) || (s.name.Contains("Particle") && mat.shader.name.Contains("Particle")))
                                {
                                    //mat = new Material(s);
                                    mat.shader = s;
                                    //SuperController.LogMessage("cubeShader " + mat.shader.name);
                                    break;
                                }
                            }
                            if (mat.shader.name == allRenderers[i].materials[j].shader.name)
                            {
                                //mat = new Material(allRenderers2[0].material.shader);
                                mat.shader = allRenderers2[0].material.shader;
                                //SuperController.LogMessage("cubeShader_CASE_StandardS " + allRenderers[i].materials[j].shader.name);
                                //mat.CopyPropertiesFromMaterial(allRenderers[i].materials[j]);
                                AssignNewShaderToMaterial(mat, allRenderers[i].materials[j].shader, mat.shader, allRenderers[i].materials[j].GetTag("RenderType", false));
                            }

                        }
                    }
                    string n2 = "R" + i + "M" + j;
                    if (mat.HasProperty("_Color")) AddColorList(mat.GetColor("_Color"), n2 + " _Color");
                    if (mat.HasProperty("_Tint")) AddColorList(mat.GetColor("_Tint"), n2 + " _Tint");
                    if (mat.HasProperty("_TintColor")) AddColorList(mat.GetColor("_TintColor"), n2 + " _TintColor");
                    if (mat.HasProperty("_EmisColor")) AddColorList(mat.GetColor("_EmisColor"), n2 + " _EmisColor");
                    if (mat.HasProperty("_EmissionColor")) AddColorList(mat.GetColor("_EmissionColor"), n2 + " _EmissionColor");


                    if (allRenderers2.Count > 0)
                    {
                        //if (allRenderers[i].materials[j].HasProperty("_TintColor")) mat.SetColor("_Color", allRenderers[i].materials[j].GetColor("_TintColor"));
                        //if (allRenderers[i].materials[j].HasProperty("_Tint")) mat.SetColor("_Color", allRenderers[i].materials[j].GetColor("_Tint"));
                    }

                    origMaterials.Add(new MatPackage(i, j, allRenderers[i].materials[j], false));
                    //if (IsValidMaterial(allRenderers[i].materials[j]))
                    newMaterials.Add(new MatPackage(i, j, new Material(mat), true));

                }
            }

            foreach (JSONStorableStringChooser chooser in jpop32List) {
                RegisterStringChooser(chooser);
                if (GetStoredNode(chooser.name, out node))
                    chooser.valNoCallback = node.Value;
            }
            foreach (JSONStorableColor colorN in ColorsList) {
                RegisterColor(colorN);
                if (GetStoredNode(colorN.name, out node)) {
                    HSVColor cN;
                    cN.H = node["h"].AsFloat;
                    cN.S = node["s"].AsFloat;
                    cN.V = node["v"].AsFloat;
                    colorN.valNoCallback = cN;
                }
            }
            foreach (JSONStorableFloat aN in AlphasList)
            {
                RegisterFloat(aN);
                if (GetStoredNode(aN.name, out node))
                    aN.valNoCallback = node.AsFloat;
            }

            // replace renders from original to copy materials
            for (int i = 0; i < allRenderers.Count; i++) {
                Material[] matsToAssign = allRenderers[i].materials;
                for (int j = 0; j < allRenderers[i].materials.Length; j++) {
                    Material vamMat = GetVamMaterialByIndex(i, j);
                    matsToAssign[j] = vamMat;
                }
                // assign new materials to prop renderer as array
                allRenderers[i].materials = matsToAssign;
            }



            if (allRenderers2.Count > 0)
            {
                List<string> shaderNames = new List<string>();
                foreach (Renderer r in allRenderers2)
                {
                    shaderNames.Add(r.material.shader.name);
                }
                //SuperController.LogMessage("ConvertMaterial allRenderers2.Count " + allRenderers2.Count);
                jpopCubeShaders = new JSONStorableStringChooser("CubeShaders", shaderNames, "None", "Cube Shaders", SetCubeShaders);
                RegisterStringChooser(jpopCubeShaders);
                if (GetStoredNode("CubeShaders", out node))
                    jpopCubeShaders.val = node.Value;

                jShadowsZWrite = new JSONStorableBool("_ShadowsZWrite", true, SetShadowsZWrite);
                RegisterBool(jShadowsZWrite);
                if (GetStoredNode("_ShadowsZWrite", out node))
                    jShadowsZWrite.val = node.AsBool;
            }

            if (GetStoredNode("StencilRef", out node))
                jpopStencilRef.valNoCallback = node.Value;
            SetpopStencilRef(jpopStencilRef.val);

            jsRenderQueue = new JSONStorableFloat("RenderQueue", 2000.0f, SetRQ, -1f, 5000f, true);
            RegisterFloat(jsRenderQueue);
            if (GetStoredNode("RenderQueue", out node))
                jsRenderQueue.val = node.AsFloat;

            foreach (JSONStorableBool aN in GOEnabList)
            {
                RegisterBool(aN);
                if (GetStoredNode(aN.name, out node))
                    aN.val = node.AsBool;
            }
            foreach (JSONStorableBool aN in RendEnabList)
            {
                RegisterBool(aN);
                if (GetStoredNode(aN.name, out node))
                    aN.val = node.AsBool;
            }

            if (jpopRenderMode.val != jpopRenderMode.defaultVal)
                SetRenderMode(jpopRenderMode.val);

            Setpop(jpop.val);

            for (int i = 0; i < jpop32List.Count; i++) {
                if (jpop32List[i].val != jpop32List[i].defaultVal) {
                    Setpop32N(jpop32List[i]);
                }
            }
            SetAnchor(jpopSelectTransform.val);
            foreach (JSONStorableColor colorN in ColorsList) SetColorN(colorN);
            foreach (JSONStorableFloat aN in AlphasList) SetAlphaN(aN);
        }

        public void AddColorList(Color c, string n)
        {
            bool a = false;
            for (int i = 0; i < AlphasList.Count; i++)
            {
                if (AlphasList[i].name == n)
                {
                    a = true;
                    break;
                }
            }
            if (!a) AlphasList.Add(new JSONStorableFloat(n + " Alpha", c.a, SetAlphaN, 0f, 10f, true));

            HSVColor colorHSVC = HSVColorPicker.RGBToHSV(c.r, c.g, c.b);
            ColorsList.Add(new JSONStorableColor(n, colorHSVC, SetColorN));
            //ColorsList[ColorsList.Count - 1].setJSONCallbackFunction += SetColorN;
        }

        public bool GetStoredNode(string storableName, out JSONNode node)
        {
            node = null;
            if (myJsnode[storableName] != null)
            {
                node = myJsnode[storableName];
                //SuperController.LogMessage("get storable: " + storableName + " value: " + myJsnode[storableName].ToString());
                return true;
            }
            return false;
        }


        protected void SetAnchor(string v)
        {
            Atom a = SuperController.singleton.GetAtomByUid(v);
            if (a == null)
            {
                SuperController.LogMessage("SetAnchor: anchor atom is null");
                return;
            }
            Transform t = a.transform;
            for (int i = 0; i < a.reParentObject.childCount; i++)
            {
                if (a.reParentObject.GetChild(i).name == "control")
                {
                    //SuperController.LogMessage(a.name + " control pos: " + a.reParentObject.GetChild(i).position);
                    t = a.reParentObject.GetChild(i);
                    break;
                }
            }
            //SuperController.LogMessage("SetAnchor: " + v + " pos: " + t.position);
            for (int i = 0; i < allRenderers.Count; i++)
            {
                allRenderers[i].probeAnchor = t;
                if (containingAtom.name == v) allRenderers[i].probeAnchor = null;
            }

        }

        protected void Setpop(string v) {
            //SuperController.LogMessage("Setpop :"+v);
            jpop.choices = null;
            jpop.choices = renderersNames;

            if (uipop32 != null) RemovePopup(uipop32);
            RemovePopup(jpopSelectTransform);
            RemovePopup(jpopStencilRef);
            RemovePopup(jpopStencilRef2);
            RemovePopup(jpopRenderMode);
            RemovePopup(jpopCubeShaders);
            for (int i = 0; i < UIColorList.Count; i++)
            {
                RemoveColorPicker(UIColorList[i]);
            }
            if (layersList.Count > 0)
            {
                for (int i = 0; i < 32; i++)
                {
                    RemoveToggle(layersList[i]);
                }
            }

            if (infoBox != null) infoBox.val = "Info\n";
            RemoveSlider(jsScaleX);
            RemoveSlider(jsScaleY);
            RemoveSlider(jsScaleZ);
            RemoveSlider(jsRenderQueue);
            if (setAlpha != null) RemoveSlider(setAlpha);
            if (uiGOEnab != null) RemoveToggle(uiGOEnab);
            if (uiRendEnab != null) RemoveToggle(uiRendEnab);
            RemoveToggle(jCastShadows);
            RemoveToggle(jReceiveShadows);
            RemoveToggle(jShadowsZWrite);
            RemoveToggle(jStopEditMaterial);

            if (v == "ALL") {
                if (infoBox != null) infoBox.val += "\nALL: changes to all objects\n\nLIGHT: first finded light object, you can change cullingMask (lighted layers)\n\nRender N Mat N: means meshes with materials, changes are individual per mesh renderer / material";
                uipop32 = CreateScrollablePopup(jpop32);
                CreateFilterablePopup(jpopSelectTransform, true);
                jpopSelectTransform.popup.onOpenPopupHandlers = OnOpenPopup1;

                CreateScrollablePopup(jpopStencilRef);
                if (haveStencilMask)
                {
                    CreateScrollablePopup(jpopStencilRef2);
                }
                else if (shadersAtom != null)
                {
                    CreateScrollablePopup(jpopStencilRef);
                }
                CreateScrollablePopup(jpopRenderMode);
                if (jpopCubeShaders != null) CreateScrollablePopup(jpopCubeShaders);
                CreateSlider(jsScaleX);
                CreateSlider(jsScaleY);
                CreateSlider(jsScaleZ);
                if (jsRenderQueue != null) CreateSlider(jsRenderQueue);
                CreateToggle(jCastShadows, true);
                CreateToggle(jReceiveShadows, true);
                if (jShadowsZWrite != null) CreateToggle(jShadowsZWrite, true);
                CreateToggle(jStopEditMaterial);
            }
            else if (v == "LIGHT") {
                for (int i = 0; i < 32; i++) {
                    CreateToggle(layersList[i], false);
                }
            }
            else
            {
                string[] ij = v.Split(' ');
                if (ij.Length == 4)
                {
                    foreach (JSONStorableBool js in GOEnabList)
                    {
                        if (js.name.EndsWith(ij[1]))
                        {
                            uiGOEnab = CreateToggle(js);
                            break;
                        }
                    }
                    foreach (JSONStorableBool js in RendEnabList)
                    {
                        if (js.name.EndsWith(ij[1]))
                        {
                            uiRendEnab = CreateToggle(js);
                            break;
                        }
                    }
                    if (allRenderers.Count > 0) {
                        Material mN = GetVamMaterialByIndex(Int32.Parse(ij[1]), Int32.Parse(ij[3]));
                        infoBox.val += "\nRenderer: " + allRenderers[Int32.Parse(ij[1])] + "\n\nMaterial: " + mN.name + "\n\nShader: " + mN.shader.name;
                        uipop32 = CreateScrollablePopup(jpop32List[Int32.Parse(ij[1])]);
                    }
                    string n = "R" + ij[1] + "M" + ij[3];
                    for (int i = 0; i < AlphasList.Count; i++)
                    {
                        if (AlphasList[i].name.Contains(n))
                        {
                            setAlpha = CreateSlider(AlphasList[i]);
                            break;
                        }
                    }
                    int i2 = 0;
                    foreach (JSONStorableColor colorN in ColorsList) {
                        if (colorN.name.Contains(n)) {
                            if (UIColorList.Count <= i2)
                            {
                                UIColorList.Add(CreateColorPicker(colorN));
                            }
                            else
                            {
                                UIColorList[i2] = CreateColorPicker(colorN);
                            }
                            i2++;
                            infoBox.val += "\n";
                            //break;
                        }
                    }
                }
            }
        }
        protected void SetStopEditMaterial(bool b)
        {
            RestoreOriginalMaterials();
            while (renderersNames.Count > 1 && renderersNames[renderersNames.Count-1].StartsWith("Render"))
            {
                renderersNames.RemoveAt(renderersNames.Count - 1);
            }
            jpop.choices = null;
            jpop.choices = renderersNames;
        }
        protected void SetGOEnable(JSONStorableBool js)
        {
            int i = int.Parse(js.name.Split(' ')[1]);
            if (allRenderers.Count > i)
            {
                allRenderers[i].gameObject.SetActive(js.val);
            }
        }
        protected void SetRenderEnable(JSONStorableBool js)
        {
            int i = int.Parse(js.name.Split(' ')[1]);
            if(allRenderers.Count > i)
            {
                allRenderers[i].enabled = js.val;
            }
        }

        protected void OnOpenPopup1()
        {
            List<string> s = new List<string>();
            foreach (Atom a in SuperController.singleton.GetAtoms())
            {
                s.Add(a.name);
            }
            List<string> options = jpopSelectTransform.choices;
            options.Clear();
            for (int e = 0; e < s.Count; ++e)
                options.Add(s[e]);
            jpopSelectTransform.choices = null; // force UI sync
            jpopSelectTransform.choices = options;
            //SuperController.LogMessage("Count: " + s.Count);
        }
        protected void OnAtomRename1(string oldid, string newid)
        {
            if(oldid == jpopSelectTransform.val) jpopSelectTransform.val = newid;
        }
        
        protected void SetCubeShaders(string v)
        {
            Shader s = null;
            foreach (Renderer r in allRenderers2)
            {
                if (r.material.shader.name == v) s = r.material.shader;
            }
            foreach (MatPackage matN in newMaterials)
            {
                if(s != null) matN.material.shader = s;
            }
        }
        protected void SetCastShadows(bool v)
        {
            foreach (Renderer r in allRenderers)
            {
                if(v) r.shadowCastingMode = ShadowCastingMode.On;
                else r.shadowCastingMode = ShadowCastingMode.Off;
            }
        }
        protected void SetReceiveShadows(bool v)
        {
            foreach (Renderer r in allRenderers)
            {
                r.receiveShadows = v;
            }
        }
        protected void SetShadowsZWrite(bool v)
        {
            foreach (MatPackage matN in newMaterials)
            {
                if (matN.material.HasProperty("_ShadowsZWrite"))
                {
                    if (v) matN.material.SetInt("_ShadowsZWrite", 1);
                    else matN.material.SetInt("_ShadowsZWrite", 0);
                }
            }
        }
        protected void SetRenderMode(string v)
        {
            foreach (MatPackage matN in newMaterials)
            {
                SetupMaterialWithBlendMode(matN.material, v, true);
            }
        }
        protected void SetRQ(float v)
        {
            foreach (MatPackage matN in newMaterials)
            {
				matN.material.renderQueue = (Int32)v;
            }
        }
        protected void SetAlphaN(JSONStorableFloat js)
        {
            string[] np = js.name.Split(' ');

            foreach (MatPackage matN in newMaterials)
            {
                if (matN.colorName == np[0])
                {
                    Color c = matN.material.GetColor(np[1]);
                    c.a = js.val;
                    matN.material.SetColor(np[1], c);
                    break;
                }
            }
        }
        protected void Setpop32(string v) {
			if(jpop32List.Count > 0 && allRenderers.Count > 0){
				for (int i = 0; i < jpop32List.Count; i++) {
					//SuperController.LogMessage("SetLayerAll: i"+i);
					jpop32List[i].val = v;
				}
			}else if(!useAllRenderers){
				for (int i = 0; i < containingAtom.childAtomContainer.childCount; i++){
					SetLayerAll(containingAtom.childAtomContainer.GetChild(i),Int32.Parse(v));
				}
			}

        }
        public void SetpopStencilRef(string v)
        {
            if (allRenderers.Count > 0)
            {
                int val = Int32.Parse(v);
                foreach (MatPackage matN in newMaterials)
                {
                    if (matN.material.HasProperty("_StencilRef"))
                    {
                        if (matN.material.shader.name == "MAG/Blackhole") continue;
                        if (haveStencilMask)
                        {
                            if (matN.material.name.StartsWith("MStencilMask (")) matN.material.SetInt("_StencilRef", Int32.Parse(jpopStencilRef.val));
                            if (matN.material.name.StartsWith("MStencilMask 1")) matN.material.SetInt("_StencilRef", Int32.Parse(jpopStencilRef2.val));
                        }
                        else
                            matN.material.SetInt("_StencilRef", val);
                    }
                }
            }
        }

        protected void Setpop32N(JSONStorableStringChooser js) {
			try {
				allRenderers[Int32.Parse(js.name.Split(' ')[1])].gameObject.layer = Int32.Parse(js.val);
			}
			catch (Exception e) {
				SuperController.LogError("DLUtils: Setpop32N: " + e);
			}
		}

		protected void SetLayerAll(Transform t,int la)
        {
			// if(allRenderers.Count > 0){
			// 	for (int i = 0; i < allRenderers.Count; i++) {
			// 		allRenderers[i].gameObject.layer = la;
			// 		SuperController.LogMessage("tName aR : " + allRenderers[i].name);
			// 	}
			
			t.gameObject.layer = la;
			//SuperController.LogMessage("tName : " + t.name);
			for (int i = 0; i < t.childCount; i++)
			{
				SetLayerAll(t.GetChild(i),la);
			}
        }

		protected void SetLayerN(JSONStorableBool jv) {
			int n = int.Parse(jv.name.Split(' ')[1]);
			//SuperController.LogMessage("SetLayerN "+n);
			//SuperController.LogMessage("SetLayerNC "+l1.cullingMask );
			int x = 0;
			for(int i=0; i<32; i++){
				if (layersList[i].val) x ^= (1 << i);
			}
			l1.cullingMask = x;
			//SuperController.LogMessage("SetLayerNC "+l1.cullingMask );
			l1CullingMask.val = (float)x;
		}

		protected Material GetVamMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in newMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		void  OnDestroy()
        {
            //AssetLoader.DoneWithAssetBundleFromFile(GetPackagePath(this) + "Custom/Assets/DemonLord/custom_subsurface_glossnmtessmappedfixedcomputebuff.assetbundle");
            if (SuperController.singleton != null)
            {
                SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename1;
            }
            rescaleObject.transform.localScale = scaleWas;
            if (l1!=null)l1.cullingMask = l1MaskWas;
			RestoreOriginalMaterials();

            if (cUAL != null)
            {
                cUAL.DeregisterAssetClearedCallback(AssetCleared);
                cUAL.DeregisterAssetLoadedCallback(AssetLoaded);
            }
            if (cUAL2 != null)
            {
                cUAL2.DeregisterAssetClearedCallback(AssetCleared2);
                cUAL2.DeregisterAssetLoadedCallback(AssetLoaded2);
            }
        }

		protected Material GetOrigMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in origMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		protected void RestoreOriginalMaterials() {
			if (origMaterials.Count == 0) return;
			for (int i = 0; i < allRenderers.Count; i++) {
                allRenderers[i].enabled = true;
                if (allRenderers[i] != null) {   // renderer may have been destroyed already by atom remove
					List<Material> matsToAssign = new List<Material>();
					for (int j = 0; j < allRenderers[i].materials.Length; j++) {
						matsToAssign.Add(GetOrigMaterialByIndex(i, j));
					}
					allRenderers[i].materials = matsToAssign.ToArray();
					//allRenderers[i].gameObject.layer = 0;
				}
			}
		}

		protected void SetColorN(JSONStorableColor jcolor) {
			string[] np = jcolor.name.Split(' ');

			foreach (MatPackage matN in newMaterials) {
				if (matN.colorName == np[0]){
                    //SuperController.LogMessage("SetColorN: " + matN.propertyName);
                    Color c = jcolor.colorPicker?.currentColor ?? HSVColorPicker.HSVToRGB(jcolor.val);
					c.a = matN.material.GetColor(np[1]).a;
                    matN.material.SetColor(np[1], c);
					
					break;
				}
			}
        }
        public void AssignNewShaderToMaterial(Material material, Shader oldShader, Shader newShader, string rTag)
        {
            int bm = -1;
            if (rTag == "Opaque" || rTag == "") bm = 0;
            if (rTag == "TransparentCutout") bm = 1;
            if (rTag == "Transparent") bm = 3;

            // _Emission property is lost after assigning Standard shader to the material
            // thus transfer it before assigning the new shader
            if (material.HasProperty("_Emission"))
            {
                material.SetColor("_EmissionColor", material.GetColor("_Emission"));
            }

            material.shader = newShader;

            if (oldShader == null || !oldShader.name.Contains("Legacy Shaders/"))
            {
                //SetupMaterialWithBlendMode(material, blendMode[(Int32)material.GetFloat("_Mode")], true);
                //return;
            }

            int bM = 0;
            if (oldShader.name.Contains("/Transparent/Cutout/"))
            {
                bM = 1;
            }
            else if (oldShader.name.Contains("/Transparent/"))
            {
                // NOTE: legacy shaders did not provide physically based transparency
                // therefore Fade mode
                bM = 3;
            }
            if (bm > 0)
                bM = bm;
            material.SetFloat("_Mode", (float)bM);

            SetupMaterialWithBlendMode(material, blendMode[bM], true);
        }
        public static void SetupMaterialWithBlendMode(Material material, string blendMode, bool overrideRenderQueue)
        {

            int minRenderQueue = -1;
            int maxRenderQueue = 5000;
            int defaultRenderQueue = -1;
            switch (blendMode)
            {
                case "Opaque":
                    material.SetFloat("_Mode", 1f);
                    material.SetOverrideTag("RenderType", "");
                    material.SetFloat("_SrcBlend", (float)UnityEngine.Rendering.BlendMode.One);
                    material.SetFloat("_DstBlend", (float)UnityEngine.Rendering.BlendMode.Zero);
                    material.SetFloat("_ZWrite", 1.0f);
                    material.DisableKeyword("_ALPHATEST_ON");
                    material.DisableKeyword("_ALPHABLEND_ON");
                    material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
                    minRenderQueue = -1;
                    maxRenderQueue = (int)UnityEngine.Rendering.RenderQueue.AlphaTest - 1;
                    defaultRenderQueue = -1;
                    break;
                case "Cutout":
                    material.SetFloat("_Mode", 2f);
                    material.SetOverrideTag("RenderType", "TransparentCutout");
                    material.SetFloat("_SrcBlend", (float)UnityEngine.Rendering.BlendMode.One);
                    material.SetFloat("_DstBlend", (float)UnityEngine.Rendering.BlendMode.Zero);
                    material.SetFloat("_ZWrite", 1.0f);
                    material.EnableKeyword("_ALPHATEST_ON");
                    material.DisableKeyword("_ALPHABLEND_ON");
                    material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
                    minRenderQueue = (int)UnityEngine.Rendering.RenderQueue.AlphaTest;
                    maxRenderQueue = (int)UnityEngine.Rendering.RenderQueue.GeometryLast;
                    defaultRenderQueue = (int)UnityEngine.Rendering.RenderQueue.AlphaTest;
                    break;
                case "Fade":
                    material.SetFloat("_Mode", 3f);
                    material.SetOverrideTag("RenderType", "Transparent");
                    material.SetFloat("_SrcBlend", (float)UnityEngine.Rendering.BlendMode.SrcAlpha);
                    material.SetFloat("_DstBlend", (float)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
                    material.SetFloat("_ZWrite", 0.0f);
                    material.DisableKeyword("_ALPHATEST_ON");
                    material.EnableKeyword("_ALPHABLEND_ON");
                    material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
                    minRenderQueue = (int)UnityEngine.Rendering.RenderQueue.GeometryLast + 1;
                    maxRenderQueue = (int)UnityEngine.Rendering.RenderQueue.Overlay - 1;
                    defaultRenderQueue = (int)UnityEngine.Rendering.RenderQueue.Transparent;
                    break;
                case "Transparent":
                    //SuperController.LogMessage("Setup :" + blendMode);
                    material.SetFloat("_Mode", 4f);
                    material.SetOverrideTag("RenderType", "Transparent");
                    material.SetFloat("_SrcBlend", (float)UnityEngine.Rendering.BlendMode.One);
                    material.SetFloat("_DstBlend", (float)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
                    material.SetFloat("_ZWrite", 0.0f);
                    material.DisableKeyword("_ALPHATEST_ON");
                    material.DisableKeyword("_ALPHABLEND_ON");
                    material.EnableKeyword("_ALPHAPREMULTIPLY_ON");
                    minRenderQueue = (int)UnityEngine.Rendering.RenderQueue.GeometryLast + 1;
                    maxRenderQueue = (int)UnityEngine.Rendering.RenderQueue.Overlay - 1;
                    defaultRenderQueue = (int)UnityEngine.Rendering.RenderQueue.Transparent;
                    break;
            }

            if (overrideRenderQueue || material.renderQueue < minRenderQueue || material.renderQueue > maxRenderQueue)
            {
                if (!overrideRenderQueue)
                    //SuperController.LogMessage("Render queue value outside of the allowed range ({" + minRenderQueue + "} - {" + maxRenderQueue + "}) for selected Blend mode, resetting render queue to default");
                material.renderQueue = defaultRenderQueue;
            }
        }

    }
}